/**************************************************************************
 * Module name:   control.c						  *
 * Project:       xps (Process State Programm for X			  *
 * written by:    Thomas Breitfeld (breitf@infotech.tu-chemnitz.de)	  *
 * last change:   11. 4. 1994                                             *
 *									  *
 * Copyright (c) 1994 Technical University of Chemnitz-Zwickau		  *
 *		      Department of Electrical Engineering		  *
 *		      Division of Data and Communication Technologies	  *
 *									  *
 * ---------------------------------------------------------------------- *
 * description:   This module contains the main xpis control functions    *
 *                like the timerset- and interrupt functions including    *
 *		  the xpis memory managment.				  *
 * ---------------------------------------------------------------------- *
 * entry points:  run_ps() 						  *
 *		  replay()						  *
 *		  adjust_timer()					  *
 *		  control_init()					  *
 *		  change_track_length()
 *		  play()						  *
 *                play_current_psi()					  *
 *                get_track_entries()					  *
 **************************************************************************/

#include<malloc.h>
#include<stdlib.h>
#include"xpis.h"
#include"control.h"
#include"pst.h"
#include"keys.h"
#include"callback.h"
#include"icon.h"


/*
 * type to reference one process state image
 */
struct psi { 
   ProcessState *pst;
   int 		pst_entries;
};

#define increment_track_pointer(p)  \
	((p == track_base+max_track_entries-1) ? p=track_base : p++)
#define decrement_track_pointer(p)  \
	((p == track_base) ? p=track_base+max_track_entries-1 : p--)

/*
 * prototypes of private functions
 */
PRIVATE void set_next_track_entry(ProcessState *pst, int pst_entries);
PRIVATE int clear_last_track_entry(void);

/*
 * variables for managing the psi recorder
 * most of this datas will be initialized in _init() function
 */
PRIVATE struct psi *track_base;	      /*base pointer to psi recorder track*/
PRIVATE struct psi *track_act;	      /*current psi*/
PRIVATE struct psi *track_last;	      /*oldest valid entry of recorder track*/
PRIVATE struct psi *play_pointer;     /*used for replaying the psi track*/
PRIVATE int track_entries;	      /*current number of entries in track*/
PRIVATE int max_track_entries;	      /*max number of entries in psi track*/
/*
 * number of slots in system process table, extracted by pst_init
 */
PRIVATE int nr_procs;


/* ---------------------------------------------------------
 * function: get_track_entries
 *           returns the number of ps images in track
 * ---------------------------------------------------------
 */
PUBLIC int get_track_entries(void)
{
   return(track_entries);
}


/* -----------------------------------------------------------------------
 * function: control_init
 *           function allocates memory for the psi track and
 *	     initialices nr_procs
 * input:    nr_entries -> number of psi entries in the recorder track
 * return:   0 if ok
 *           -1 if malloc failed
 * -----------------------------------------------------------------------
 */
PUBLIC int control_init(int nr_entries, int nproc)
{
   /*
    * init recorder track
    */
   if(nr_entries >= 2) {
      if((track_base = (struct psi *)malloc(nr_entries*sizeof(struct psi)))
	 == NULL)
	 return(-1);
      track_act = play_pointer = track_base;
      track_last = track_base+1;
      max_track_entries = nr_entries;
      track_entries = 0;
   } else {
      puterr("control_init: recorder track entries has to be 2 or greater");
      return(-1);
   }

   /*
    * set nr_procs
    */
   nr_procs = nproc;

   return(0);
}


/* ------------------------------------------------------------------------
 * function: change_track_length()
 *           changes the number of psi entries in the recorder track
 * input:    nr_entries -> new entries number
 * return:   0 if ok
 *           -1 if failed
 * ------------------------------------------------------------------------
 */
PUBLIC int change_track_length(int nr_entries)
{
   struct psi *new_base, *pointer;
   int i, diff;
   
   if(nr_entries >= 2) {
      if(nr_entries != max_track_entries) {
	 /*
	  * allocate new psi table
	  */
	 if((new_base = (struct psi *)malloc(nr_entries*sizeof(struct psi)))
	    == NULL)
	    return(-1);

	 /*
	  * clear last track entries, if there is not sufficient space
	  * in the new track for all entries of the old track !
	  */
	 if((diff = track_entries-nr_entries) > 0)
	    for(i=0; i<diff; i++)
	    clear_last_track_entry();
	 /*
	  * copy all entries of the old track into the new track
	  */
	 pointer = new_base-1;
	 for(i=0;i<track_entries;i++) {
	    *(++pointer) = *track_last;
	    increment_track_pointer(track_last);
	 }

	 /*
	  * give off the old memory
	  */
	 free(track_base);

	 /*
	  * set pointers to the new memory;
	  */
	 track_act = play_pointer = pointer;
	 track_base = track_last = new_base;
	 max_track_entries = nr_entries;
      }
   }
   else {
      puterr("control_init: number of recorder track entries has to be \n"
	     "              two or more\n");
      return(-1);
   }
   return(0);
}


/* --------------------------------------------------------------------
 * function: play_current_psi()
 *           takes last played psi (play pointer) and 
 *	     plays it again.
 * --------------------------------------------------------------------
 */
PUBLIC void play_current_psi(void)
{
   if(track_entries > 0) {
      actualize_painting(play_pointer->pst, play_pointer->pst_entries);
   } else
      puterr("play_current_psi: there is not any psi\n");
}


/* ---------------------------------------------------------------------
 * function: set_next_track_entry
 * input:    pst -> pointer to the process state table to store
 *           pst_entries
 * ---------------------------------------------------------------------
 */
PRIVATE void set_next_track_entry(ProcessState *pst, int pst_entries)
{
   increment_track_pointer(track_act);
   if(track_act == track_last)
      clear_last_track_entry();
   /*write new entry*/
   track_act->pst = pst;
   track_act->pst_entries = pst_entries;
   track_entries++;
   play_pointer = track_act;
}


/* ------------------------------------------------------------------------
 * function: clear_last_track_entry
 * return:   0 if ok
 *           -1 if there is nothing to clear
 *           NOTE: the actual entry can never be the last entry, so it can
 *                 never be cleared
 * ------------------------------------------------------------------------
 */
PRIVATE int clear_last_track_entry(void)
{
   struct psi *latch;

   if(track_entries > 1) {
      latch = track_last;
      increment_track_pointer(track_last);
      track_entries--;
      free(latch->pst);
      return(0);
   }else
      return(-1);
}


/* --------------------------------------------------------------------------
 * function: play
 *           The play function is the main control function for replaying
 *           the record. The function changes the play pointer in accordance
 *           to the input parameters and makes the painting and the 
 *           panel items up to date with the new position
 * input:    move -> symbolic constant for the play pointer motion
 *           n -> replacment for move=ACT_N :the play pointer will be 
 *	       	  moved to the position track_act - n ;
 *           set_slider -> YES or NO:slider should be set by the function
 *				     or not
 * return:   0 -> ok
 *           1 -> play_pointer has act position
 *           -1 -> play_pointer has last position
 * --------------------------------------------------------------------------
 */
PUBLIC int play(int move, int n, bool set_slider)
{
   static int return_value = 1;
   int i, slider_value=0;
   Panel_item slider, dec_button, inc_button;
   Frame control_frame = xv_get(get_base_frame(),
				XV_KEY_DATA,CONTROL_FRAME_KEY);

   /*get panel items */
   slider = xv_get(control_frame,XV_KEY_DATA,SLIDER_CONTROL_KEY);
   inc_button = xv_get(control_frame,XV_KEY_DATA,FF_CONTROL_KEY);
   dec_button = xv_get(control_frame,XV_KEY_DATA,FB_CONTROL_KEY);
   if(set_slider == YES)
      slider_value = xv_get(slider, PANEL_VALUE);

   /*set play pointer*/
   switch(move) {

      case INCREMENT:
	 if(play_pointer != track_act) {
	    increment_track_pointer(play_pointer);
	    slider_value++;
	 }
	 break;

      case DECREMENT:
	 if(play_pointer != track_last) {
	    decrement_track_pointer(play_pointer);
	    slider_value--;
	 }
	 break;

      case LAST:
	 play_pointer = track_last;
	 slider_value = -(track_entries-1);
	 break;

      case ACT_N:
	 if(n >= track_entries)
	    n = track_entries-1;
	 play_pointer = track_act;
	 for(i=0; i<n; i++)
	    decrement_track_pointer(play_pointer);
	 slider_value = -n;
	 break;

      default:
	 puterr("play: unkown move type\n");
   }

   /*set slider */
   if(set_slider == YES) 
      xv_set(slider,PANEL_VALUE,slider_value,NULL);

   /*set inc- and dec-buttons and return value*/
   if(return_value != 0) {
      if(return_value == -1)
	 xv_set(dec_button,PANEL_INACTIVE,FALSE,NULL);
      else
	 xv_set(inc_button,PANEL_INACTIVE,FALSE,NULL);
   }
   if(play_pointer == track_last) {
      return_value = -1;
      xv_set(dec_button,PANEL_INACTIVE,TRUE,NULL);
   } else {
      if(play_pointer == track_act) {
	 return_value = 1;
	 xv_set(inc_button,PANEL_INACTIVE,TRUE,NULL);
      } else
	 return_value = 0;
   }

   /* set painting*/
   actualize_painting(play_pointer->pst, play_pointer->pst_entries);

   return(return_value);
}


/* --------------------------------------------------------------
 * function: run_ps();
 *           This function will be called by the notifier as
 *	     the result of timer signal.
 * --------------------------------------------------------------
 */
PUBLIC Notify_value run_ps(Notify_client client,int pid,
			     union wait *status,struct rusage *rusage)
{  
   ProcessState *pst;
   int pst_entries;

      /*allocate memory for pst*/
      if((pst = (ProcessState *)malloc(nr_procs * sizeof(ProcessState)))
	  == NULL) {
	 puterr("ps_ready: malloc failed ! Trying to get same memory !\n");
	 do {
	    if(clear_last_track_entry() < 0) {
	       puterr("ps_ready: there is no more memory available\n"
		      "P R O G R A M M   T E R M I N A T E D !!\n");
	       exit(1);
	    }
	 } while((pst = (ProcessState *)malloc(nr_procs*sizeof(ProcessState)))
		 == NULL);
	 fprintf(stderr,"ps_ready: got memory, but now there are only %d"
	 "pst images\n         for replaying in memory\n",track_entries);
      }

      /*write new values into pst */
      if((pst_entries = getpst(pst)) > 0) {
	 if((pst = (ProcessState *)realloc(pst,
			      pst_entries*sizeof(ProcessState))) == NULL) {
	    puterr("ps_ready: realloc failed !! actually impossible !\n"
		   "PROGRAMM  TERMINATED\n");
	    exit(1);
	 }
	 set_next_track_entry(pst, pst_entries);
	 actualize_painting(pst, pst_entries);
      }
      else {
	 puterr("ps_ready:return value of getpst=0, can't paint anything\n");
	 free(pst);
      }

      return NOTIFY_DONE;
}


/* ---------------------------------------------------------------------
 * function: replay
 *           interrupt service routine for replay timer
 * ---------------------------------------------------------------------
 */
PUBLIC Notify_value replay(void)
{
   if(play(INCREMENT, 0, YES) > 0)
      /* press stop button */
      stop_button_control_proc(
         xv_get( xv_get(get_base_frame(), XV_KEY_DATA,CONTROL_FRAME_KEY),
		 XV_KEY_DATA, REPLAY_CONTROL_KEY),
         (Event *)0);

   return NOTIFY_DONE;
}


/* -----------------------------------------------------------------------
 * function: adjust_timer
 *           sets and starts the timer for various callback functions
 * input:    sec -> seconds of timervalue
 *           sec10 -> one tenth of seconds of timervalue
 *             if both values are equal zero, the timer will be stoped
 *           cb_func -> symbolic constant to select an notify function
 * return:   NULL if sec10 > 9 or negative input values
 *	     otherwise a pointer to the timer_state struct wich 
 *           describes the time and the callback function of the 
 *           just cleared timer setting
 * -----------------------------------------------------------------------
 */
PUBLIC struct timer_state *adjust_timer(int sec, int sec10, char cb_func)
{
   static struct itimerval timer;
   static struct timer_state tstate_memory;
   static struct timer_state tstate_latch;
   Notify_value (*cbfp)();		/*pointer to callback function*/

   if((sec10 > 9) || (sec10 < 0) || (sec < 0))
      return(NULL);

   if((sec > 0) || (sec10 > 0)) { 
      switch(cb_func) {
	 case RUN_PS:
	    cbfp = run_ps;
	    break;
	 case REPLAY:
	    cbfp = replay;
	    break;
         case ANIMATE:
	    cbfp = animate;
	    break;
	 default:
	    puterr("adjust_timer: unknown callback function\n");
	    return(NULL);
      }

      timer.it_value.tv_sec = 0;
      timer.it_value.tv_usec = 10000;
      timer.it_interval.tv_sec = sec;
      timer.it_interval.tv_usec = sec10 * 100000;
      notify_set_itimer_func(get_base_frame(), cbfp,
			     ITIMER_REAL, &timer, NULL);
   } 
   else {
      /* turn timer off */
      notify_set_itimer_func(get_base_frame(), NOTIFY_FUNC_NULL,
			     ITIMER_REAL, NULL, NULL);
      cb_func = 0;
   }

   /*save new values in tstate_memory*/
   tstate_latch = tstate_memory;
   tstate_memory.sec = sec;
   tstate_memory.sec10 = sec10;
   tstate_memory.cb_func = cb_func;

   return(&tstate_latch);
}

