/**************************************************************************
 * Module name:   filter.c						  *
 * Project:       xps (Process State Programm for X			  *
 * written by:    Thomas Breitfeld (breitf@infotech.tu-chemnitz.de)	  *
 * last change:   11. 4. 1994                                             *
 *									  *
 * Copyright (c) 1994 Technical University of Chemnitz-Zwickau		  *
 *		      Department of Electrical Engineering		  *
 *		      Division of Data and Communication Technologies	  *
 *									  *
 * ---------------------------------------------------------------------- *
 * description:   input : pst                                             *
 *		  output: pdt						  *
 * ---------------------------------------------------------------------- *
 * entry point:   filter()						  *
 *                initfp()						  *
 *                text_to_fttys()					  *
 *                text_to_fuids()                                         *
 *		  pval_to_fstate()                                        *
 **************************************************************************/

#include<stdio.h>
#include<string.h>
#include<stdlib.h>

#include"xpis.h"
#include"filter.h"
#include"pst.h"
#include"paint.h"
#include"uname.h"
#include"misc.h"

typedef int (*fcmp) (const void *,const void *);
struct Help {
   int pid;
   int ppid;
   ProcessState *pstate;
};

/* function prototypes */
PRIVATE int uid_filter(Fp *fp, ProcessState *proc);
PRIVATE int tty_filter(Fp *fp, ProcessState *proc);
PRIVATE int state_filter(Fp *fp, ProcessState *proc);
PRIVATE int qcompid(ProcessToDraw *elm1,ProcessToDraw *elm2);
PRIVATE int qhelppid(struct Help *elm1,struct Help *elm2);
PRIVATE int shelpid(int *key,struct Help *elm2);
PRIVATE int shelppid(int *key,struct Help *elm2);
PRIVATE int do_stack(int pid);
#define pop()	  do_stack(-1)
#define push(pid) do_stack(pid)

/* variables for the soft stack and the help field used in filter()
 * the values will be initialized in filter_init()
 */
PRIVATE int *stack;
PRIVATE int size_of_stack;

PRIVATE struct Help *help;
PRIVATE int size_of_help;


/* ---------------------------------------------------------
 * function: filter_init
 *           initialices stack and filterparameters
 *	     look for initfp to understand the parameters
 * return:   0 if ok
 *	     -1 if malloc failed
 * ---------------------------------------------------------
 */
PUBLIC int filter_init(int nproc, int *uids, char **ttys, States *states)
{
   /* init stack */
   size_of_stack = nproc;
   if((stack = (int *)malloc(size_of_stack * sizeof(int))) == NULL)
      return(-1);

   /* init help field for filter function */
   size_of_help = nproc + 2;
   if((help = (struct Help *)malloc(size_of_help * sizeof(struct Help)))
	      == NULL)
      return(-1);

   /* init filter parameters */
   initfp(uids, ttys, states);

   return(0);
}

/* ---------------------------------------
 * function: filter
 *           look above
 * ---------------------------------------
 */
PUBLIC int filter(Fp *fp,ProcessState pst[],int pst_entries,
		   ProcessToDraw pdt[])
{
   int i,pid,k=0;
   struct Help *p;

   pdt[0].pid = -1;
   pdt[0].ppid = -1;

   /*check if show-childes-mode is active*/
   if((pid = get_childs()) < 0) {     /*inactive*/
      /*transfer all non filtered entries direct into pdt*/
      for(i=0;i<pst_entries;i++) {
	 if(uid_filter(fp,&pst[i]) 
		       && tty_filter(fp,&pst[i]) 
		       && state_filter(fp,&pst[i])) {
	    k++;
	    pdt[k].pid = pst[i].pid;
	    pdt[k].ppid = pst[i].ppid;
	    pdt[k].pstate = &pst[i];
	 }
      }
   } else {                            /*show-childs-mode active*/
      /* init help */
      memset((char *)help, -1, size_of_help * sizeof(struct Help));
      /* transfer all non filtered entries into help*/
      for(i=0;i<pst_entries;i++) {
	 if(uid_filter(fp,&pst[i]) 
		       && tty_filter(fp,&pst[i]) 
		       && state_filter(fp,&pst[i])) {
	    k++;
	    help[k].pid = pst[i].pid;
	    help[k].ppid = pst[i].ppid;
	    help[k].pstate = &pst[i];
	 } 
      }
      help[k+1].pid = -1; /*end of list sign*/
      /* sort help with the criteria of ppid */
      qsort(&help[1],k,sizeof(struct Help),(fcmp)qhelppid);
      if((p = (struct Help *)lfind(&pid, &help[1], (size_t *)&k,
	      sizeof(struct Help),(fcmp)shelpid)) != NULL) {
	 pdt[i=1].pid = p->pid;
	 pdt[1].ppid = p->ppid;
	 pdt[1].pstate = p->pstate;
	 do {
	    if((p=(struct Help *)bsearch(&pid,&help[1],k,
	       sizeof(struct Help),(fcmp)shelppid)) != NULL) {
	       while((p-1)->ppid == pid)
		  p--;
	       do {
		  if(push(pdt[++i].pid = p->pid) < 0)
		     puterr("filter: warning: stack is full !!");
		  pdt[i].ppid = p->ppid;
		  pdt[i].pstate = p->pstate;
	       } while((++p)->ppid == pid);
	    }
	 } while((pid = pop()) >= 0);
	 k = i;
      } else {
	 clear_childs();
	 for(i=1;i<=k;i++) {
	    pdt[i].pid = help[i].pid;
	    pdt[i].ppid = help[i].ppid;
	    pdt[i].pstate = help[i].pstate;
	 }
      }
   }

   /* sort pdt with the criteria of pid*/
   qsort(&pdt[1],k,sizeof(ProcessToDraw),(fcmp)qcompid);

   /*
    * the swapper process (pid = 0) is the primary process in the unix
    * system. Its ppid number is 0 too. This is an inconsistancy of the 
    * unix process tree. Set swapper's ppid to the virtual root node pdt[0]
    * if the swapper process is present in the list.
    */
    if(pdt[1].pid == 0)
       pdt[1].ppid = -1;

    return(k);
}


/* ----------------------------------------------------------
 * function: qcompid, qhelppid, shelpid, shelppid
 *	     compare functions for qsort ,bsearch and lfind
 * ----------------------------------------------------------
 */
PRIVATE int qcompid(ProcessToDraw *elm1,ProcessToDraw *elm2)
{   return(elm1->pid - elm2->pid);}

PRIVATE int qhelppid(struct Help *elm1,struct Help *elm2)
{   return(elm1->ppid - elm2->ppid);}

PRIVATE int shelpid(int *key,struct Help *elm2)
{   return(*key - elm2->pid);}

PRIVATE int shelppid(int *key,struct Help *elm2)
{   return(*key - elm2->ppid);}


/* ----------------------------------------------------------
 * function: do_stack
 *           stores int pids using the LIFO princip
 * input:    pid -> ineger to store (push)
 *                  if pid < 0 :restore one value (pop)
 * return:   push: 0 is ok
 *                 -1 is stack full
 *           pop:  stored integer if ok
 *                 -1 if stack is empty
 * ----------------------------------------------------------
 */
PRIVATE int do_stack(int pid)
{
   static int sp = -1;

   if(pid < 0) {
      if(sp < 0)
	 return(-1);
      return(stack[sp--]);
   } else {
      if((sp+1) >= size_of_stack)
	 return(-1);
      stack[++sp] = pid;
      return(0);
   }
}


/* -----------------------------------------------------------
 * function: tty_filter
 *           takes tty filtering
 * input:    proc-> pointer to one entry of pst
 *           fp -> pointer tp filter parameters
 * return:   true (1) if entry should be displayed
 *           false (0) if entry should be filterd out
 * -----------------------------------------------------------
 */
PRIVATE int tty_filter(Fp *fp, ProcessState *proc)
{
   int i = 0;
   char **ttys = fp->ttys;

   if(ttys == 0) 
      return(YES);
   else {
      while(ttys[i] != NULL) {
	 if(strcmp(proc->tty,ttys[i]) == 0)
	    return(NO);
	 i++;
      }
      return(YES);
   }
}


/* -----------------------------------------------------------
 * function: uid_filter
 *           takes uid filtering
 * input:    proc-> pointer to one entry of pst
 *           fp -> pointer to filter parameters
 * return:   true (1) if entry should be displayed
 *           false (0) if entry should be filterd out
 * -----------------------------------------------------------
 */
PRIVATE int uid_filter(Fp *fp, ProcessState *proc)
{
   int i = 0;
   int *uids = fp->uids;

   if(uids == 0) 
      return(YES);
   else {
      while(uids[i] != END_F_UIDS_LIST) {
	 if(proc->ruid == uids[i])
	    return(YES);
	 i++;
      }
      return(NO);
   }
}


/* -----------------------------------------------------------
 * function: state_filter
 *           takes process state filtering
 * input:    proc-> pointer to one entry of pst
 *           fp -> pointer to filter parameters
 * return:   true (1) if entry should be displayed
 *           false (0) if entry should be filterd out
 * -----------------------------------------------------------
 */
PRIVATE int state_filter(Fp *fp, ProcessState *proc)
{
   int i = 0;
   States *states = fp->states;

   if(states == 0) 
      return(YES);
   else {
      while(states->letter_2cd[i] != (char)0) {
	 if(proc->status[1] == states->letter_2cd[i])
	    return(YES);
	 i++;
      }
      i=0;
      while(states->letter_1st[i] != (char)0) {
	 if(proc->status[0] == states->letter_1st[i])
	    return(YES);
	 i++;
      }
      return(NO);
   }
}


/* ------------------------------------------------------------------------
 * function: text_to_fuids
 *           transforms a text string of users in a list of integers
 *           to be read by initfp
 * input:    text -> pointer to text string
 *           uids -> pointer to array of int the result will be written to
 * return:   -1 if error
 *           0  error: not all words readable
 * ------------------------------------------------------------------------
 */
PUBLIC int text_to_fuids(char *text, int *uids)
{
   int i,k=0;
   char latch[10];

   while(YES) {
      if((*text!=' ') && (*text!=',') && (*text!=(char)0)) {
	 i = 0;
	 do {
	    latch[i] = *text;
	    text++;
	    if(++i == 10) {
	       puterr("text_to_fuids:dimension of latch too short");
	       return(-1);
	    }
	 } while((*text!=' ') && (*text!=',') && (*text!=(char)0));
	 latch[i] = 0;
	 if((uids[k]=userid(latch)) < 0) {
	    uids[k] = END_F_UIDS_LIST;
	    return(0);
	 }
	 k++;
      }
      if(*text == (char)0) {
	 uids[k] = END_F_UIDS_LIST;
	 return(1);
      }
      text++;
   }
}


/* ------------------------------------------------------------------------
 * function: text_to_ttys
 *           transforms a text string of ttys in a list of pointers to 
 *           strings to be read by initfp
 * input:    text -> pointer to text string
 *           ttys -> pointer to array of pointers the result will be 
 *                   written to
 *           s_length -> max allowed length of tty strings
 * return:   -1 if error
 *           0  error: not all words readable
 * ------------------------------------------------------------------------
 */
PUBLIC int text_to_fttys(char *text,char **ttys,int s_length)
{
   int i,k=0;
   char latch[10];

   while(YES) {
      if((*text!=' ') && (*text!=',') && (*text!=(char)0)) {
	 i = 0;
	 do {
	    latch[i] = *text;
	    text++;
	    if(++i == 10) {
	       puterr("text_to_fuids:dimension of latch too short");
	       return(-1);
	    }
	 } while((*text!=' ') && (*text!=',') && (*text!=(char)0));
	 latch[i] = 0;
	 strncpy(ttys[k],latch,s_length);
	 ttys[k][s_length] = 0;
	 k++;
      }
      if(*text == (char)0) {
	 ttys[k] = NULL;
	 return(1);
      }
      text++;
   }
}


/* ------------------------------------------------------------------------
 * function: pval_to_fstate
 *           transforms the state panel value in the state structure 
 *           to be read by initfp
 * input:    text -> pointer to text string
 *           ttys -> pointer to array of pointers the result will be 
 *                   written to
 *           s_length -> max allowed length of tty strings
 * return:   -1 if error
 *           0  error: not all words readable
 * ------------------------------------------------------------------------
 */
PUBLIC int pval_to_fstate(int pval,States *states)
{
    states->letter_1st[0] = 0;
    states->letter_2cd[0] = 0;

    if(pval & S_RUNING)
       strcat(states->letter_1st,"R");
    if(pval & S_SLEEPING)
       strcat(states->letter_1st,"S");
    if(pval & S_IDLEING)
       strcat(states->letter_1st,"I");
    if(pval & S_SWAPED)
       strcat(states->letter_2cd,"W");
    if(pval & S_STOPED)
       strcat(states->letter_1st,"T");
    if(pval & S_WAITING)
       strcat(states->letter_1st,"PD");
    return(1);
}


/* ------------------------------------------------------------------------
 * function: initfp
 *	     does handling of fp structure storing the
 *           filter parameters
 * input:    uids -> pointer to array of int containing uids which should
 *                   be not filtered out. The list must be teerminated with
 *                   END_F_UIDS_LIST. If uids points to NULL no change 
 *                   is done in fp.
 *           ttys -> pointer to array of pointers to characters. 
 *                   The pointer list must be terminated with a pointer to
 *                   NULL. ttys = NULL indicates no change
 *           states->pointer to States structure used for filtering
 *                   NULL indicates no change
 * return:   pointer to fp structure
 * ------------------------------------------------------------------------
 */
PUBLIC Fp *initfp(int *uids,char **ttys,States *states)
{
   #define NUMBER_F_UIDS 20
   #define NUMBER_F_TTYS 20
   #define TTY_SIGN_LENGTH 2

   static int fuids[NUMBER_F_UIDS] = { END_F_UIDS_LIST };
   static char fttytext[NUMBER_F_TTYS][TTY_SIGN_LENGTH+1];
   static char *fttys[NUMBER_F_TTYS] = {NULL};
   static States fstates = {{'\0'},{'\0'}};
   static Fp fp = {NULL,NULL,NULL};

   int i=0;

   /* set uids
    */
   if(uids != NULL) {
      if(*uids >= 0) {
	 do {
	    fuids[i] = uids[i];
	    if(++i == (NUMBER_F_UIDS - 1)) {
	       puterr("initfp: uids list termination not found within the\n"
		   "        expected size, uids list will be truncated\n"); 
	       break;
	    }
	 } while(uids[i] >= 0);
	 fuids[i] = END_F_UIDS_LIST;
	 fp.uids = fuids;
      }
      else 
	 fp.uids = NULL;
   }

   /* set ttys
    */
   i=0;
   if(ttys != NULL) {
      if(*ttys != NULL) {
	 do {	    
	    fttys[i] = strncpy(fttytext[i], ttys[i], TTY_SIGN_LENGTH);
	    fttytext[i][TTY_SIGN_LENGTH] = 0;
	    if(++i == (NUMBER_F_TTYS - 1)) {
	       puterr("initfp: ttys list termination not found within the\n"
		    "        expected size, fttys list will be truncated\n"); 
	       break;
	    }
	 } while(ttys[i] != NULL);
	 fttys[i] = NULL;
	 fp.ttys = fttys;
      }
      else
	 fp.ttys = NULL;
   }

   /* set states 
    */
   i=0;
   if(states != NULL) {
      if((*(states->letter_1st) != 0) || (*(states->letter_2cd) != 0)) {
	 while(states->letter_1st[i] != 0) {
	    fstates.letter_1st[i] = states->letter_1st[i];
	    if(++i == (NUMBER_1ST_LETTERS - 1)) {
	       puterr("initfp: 1st_letters list termination not found within"
	       "the\n   expected size, 1st_letters list will be truncateded\n"); 
	       break;
	    }
	 }
	 fstates.letter_1st[i] = 0;
	 i=0;
	 while(states->letter_2cd[i] != 0) {
	    fstates.letter_2cd[i] = states->letter_2cd[i];
	    if(++i == (NUMBER_2CD_LETTERS - 1)) {
	       puterr("initfp: 2cd_letters list termination not found within"
	       "the\n   expected size, 2cd_letters list will be truncateded\n"); 
	       break;
	    }
	 }
	 fstates.letter_2cd[i] = 0;
	 fp.states = &fstates;
      }
      else
	 fp.states = NULL;
   }

   return(&fp); 
}      

