/**************************************************************************
 * Module name:   paint.c			                          *
 * Project:       xps (Process State Programm for X			  *
 * written by:    Thomas Breitfeld (breitf@infotech.tu-chemnitz.de)	  *
 * last change:   11. 4. 1994                                             *
 *									  *
 * Copyright (c) 1994 Technical University of Chemnitz-Zwickau		  *
 *		      Department of Electrical Engineering		  *
 *		      Division of Data and Communication Technologies	  *
 *									  *
 * ---------------------------------------------------------------------- *
 * description:   This module contains some functions to paint		  *
 *                the canvas				       	          *
 *                and the definition of pdt and pst.	                  *
 * ---------------------------------------------------------------------- *
 * entry points:   actualize_painting()					  *
 *		   paint()						  *
 *                 repaint()						  *
 *                 pdt_search						  *
 *                 item_select						  *
 *		   paint_object						  *
 **************************************************************************/

#include<stdio.h>
#include<stdlib.h>

#include"xpis.h"
#include"paint.h"
#include"filter.h"
#include"misc.h"
#include"xpsgc.h"
#include"prop.h"

#define MAX_LEVEL 	20 /* max number of tree levels */

#define POINT_IN_RECT(px,py,rx,ry,rw,rh) \
			((px) >= rx && (py) >= ry && \
			(px) < rx+rw && (py) < ry+rh)

typedef int (*fcmp)(const void *,const void *);

/*function prototypes */
PRIVATE int  qscomp(int *key,ProcessToDraw *elm);
PRIVATE void make_btree(void);
PRIVATE void compute_list_pos(int node,int xdim,int *x,int *y,XpisGC *gc);
PRIVATE void compute_tree_pos(int *xdim,int *ydim,XpisGC *gc);
PRIVATE int  compute_tree(int node,int minx[],int *maxlevel,XpisGC *gc);
PRIVATE void shift_node(int node,int pixels, int minx[], int level,XpisGC *gc);
PRIVATE void set_process_icon(ProcessToDraw *proc, XpisGC *gc);
PRIVATE void set_image(ProcessToDraw *proc, XpisGC *gc);
PRIVATE void paint_list(XpisGC *gc);
PRIVATE void paint_for_list(int node,XpisGC *gc);
PRIVATE void paint_for_tree(int node,XpisGC *gc);
PRIVATE void paint_all_trees(XpisGC *gc);


/*pdt declaration, pdt is the database for the repaint proc*/
PRIVATE ProcessToDraw *pdt1;
PRIVATE ProcessToDraw *pdt2;
PRIVATE ProcessToDraw *pdt;
PRIVATE pdt_entries = 0;


/* -----------------------------------------------------------
 * function: paint_init
 *           allocates memory for the two pdt's
 * return:   0 if ok
 *           -1 if malloc failed
 * -----------------------------------------------------------
 */
PUBLIC int paint_init(int nproc)
{
   if((pdt1 = (ProcessToDraw *)malloc((nproc+1) * sizeof(ProcessToDraw)))
	      == NULL)
      return(-1);

   if((pdt2 = (ProcessToDraw *)malloc((nproc+1) * sizeof(ProcessToDraw)))
	      == NULL)
      return(-1);

   pdt = pdt1;

   return(0);
}


/* --------------------------------------------------------------
 * function : pdt_search()
 * --------------------------------------------------------------
 */
PUBLIC ProcessToDraw *pdt_search(int pid)
{
   return((ProcessToDraw *)bsearch(&pid,&pdt[1],pdt_entries,
	  sizeof(ProcessToDraw),(fcmp)qscomp));
}

 
/* --------------------------------------------------------------
 * function: actualize_painting()
 *	     This function filters the pst to the pdt and handles 
 *	     the further steps to redesign the painting if 
 *	     something has been changed
 *
 * --------------------------------------------------------------
 */
PUBLIC void actualize_painting(ProcessState *pst, int pst_entries)
{
   static int pd_table = 1;
   ProcessToDraw *new_pdt = (pd_table == 1) ? pdt2 : pdt1;
   ProcessToDraw *proc, *new_proc;
   int i, old_entries = pdt_entries;
   XpisGC *gc = getgc();

   if((pdt_entries=filter(getfp(),pst,pst_entries,new_pdt)) != old_entries) {
      pdt = new_pdt;
      pd_table = (pd_table == 1) ? 2 : 1;
      paint(gc);
      set_pitems_s();
      return;
   }

   for(i=pdt_entries;i>0;i--) {
      if((pdt[i].pid!=new_pdt[i].pid) || (pdt[i].ppid!=new_pdt[i].ppid)) {
	 pdt = new_pdt;
	 pd_table = (pd_table == 1) ? 2 : 1;
	 paint(gc);
	 set_pitems_s();
	 return;
      }
   }

   for(i=1;i<=pdt_entries;i++) {
      proc = &pdt[i];
      new_proc = &new_pdt[i];
      if(*(proc->pstate->status) != *(new_proc->pstate->status)) {
	 proc->pstate = new_proc->pstate;
	 proc->picon.text = new_proc->pstate->text;
	 set_process_icon(proc, gc);
	 paint_object(proc->picon,gc);
      }
      else {
	 proc->pstate = new_proc->pstate;
	 proc->picon.text = new_proc->pstate->text;
      }
   }
   set_pitems_s();
}


/* ------------------------------------------------------------------
 * function: paint()
 *           paint() computes the positions of the process Icons in 
 *           accordance to the xpisgc new and starts the physical 
 *           painting as well.
 * input:    gc
 *           global pdt 
 * ------------------------------------------------------------------
 */
PUBLIC void paint(XpisGC *gc)
{
   int i, ydim=0,xdim=0, x=0, y=0;

   pdt[0].picon.w = 0;         /*set width of dummy root process*/
   for(i=1;i<=pdt_entries;i++) /*set icon properties of all others*/
      set_process_icon(&pdt[i],gc);

   make_btree();

   if(gc->tree == YES) {
      compute_tree_pos(&xdim,&ydim,gc);
   }
   else {
      xdim = xv_get(gc->canvas,XV_WIDTH) - 30;
      compute_list_pos(0, xdim, &x, &y, gc);
      ydim = y + gc->IconHeight + gc->VerSpareSpace;
   }

/*   printf("paint: xdim=%d  ydim=%d\n",xdim,ydim); */
   xv_set(gc->canvas,
	CANVAS_HEIGHT,ydim,
	CANVAS_WIDTH,xdim,
	NULL);

   XClearWindow(gc->dpy, gc->xwin);

   /* 
    *paint the new state of the art to the canvas 
    */
   repaint((Canvas)0,(Xv_Window)0,(Rectlist *)0);
}


/* -------------------------------------------------------------------
 * function: repaint
 *	     repaint procedure for canvas paint window
 *           it will be called by the Xview notifier
 * -------------------------------------------------------------------
 */
PUBLIC void repaint(Canvas canvas,Xv_Window paint_window,Rectlist *rects)
{
   XpisGC *gc = getgc();

   if(gc->tree == YES)
      paint_all_trees(gc);
   else
      paint_list(gc);
}


/* -------------------------------------------------------------------
 * function: paint_all_trees
 *	     This function loops through all root nodes of the 
 *	     existing trees and paint them  side by side.
 *	     This function is necassary because there is the
 *	     possibility of the presence of more then one
 *	     root node because of the filtering. The first entry
 *	     of the pdt contains the pointer to the list of root nodes
 * input:    global pdt
 *           pointer to xpisgc
 * ------------------------------------------------------------------
 */
PRIVATE void paint_all_trees(XpisGC *gc)
{
   int p;

   if(pdt_entries > 0) {
      p = pdt[0].cp;
      while(p != 0) {
	 paint_for_tree(p,gc);
	 p = pdt[p].rp;
      }
   }
   /*
    *else
    *  puterr("paint_all_trees: there is no tree to paint\n");
    */
}


/* ------------------------------------------------------------------
 * function: paint_for_tree
 *           paints one process and its childs stored in pdt 
 *           connected with lines to show a tree
 * input:    node is the index number of an pdt entry .
 *                This entry and all its subentries will be painted
 *           pointer to xpigc graphic context
 *	     global pdt
 * ------------------------------------------------------------------
 */
PRIVATE void paint_for_tree(int node,XpisGC *gc)
{
   int p;
   ProcessIcon *i1,*i2;

   i1 = &pdt[node].picon;
   paint_object(*i1,gc);
   if((p = pdt[node].cp) != 0) {
      do {
	 i2 = &pdt[p].picon;
	 XDrawLine(gc->dpy,gc->xwin,gc->gc,
		   i1->x + (i1->w / 2),	/* x1 */
		   i1->y + i1->h,	/* y1 */
		   i2->x + (i2->w / 2),	/* x2 */
		   i2->y);		/* y2 */
	 paint_for_tree(p,gc);
      } while((p=pdt[p].rp) != 0);
   }
}


/* -------------------------------------------------------------------
 * function: paint_list
 *           paints all root processe and their childs at the position 
 *           given in the picon structures of a pdt entry
 * input :   global pdt
 *           gc -> pointer to xpisgc
 * -------------------------------------------------------------------
 */
PRIVATE void paint_list(XpisGC *gc)
{
   int p;

   if(pdt_entries > 0) {
      p = pdt[0].cp;
      while(p != 0) {
	 paint_for_list(p,gc);
	 p = pdt[p].rp;
      }
   }
   else
      puterr("paint_list: there is no list to paint\n");
}


/* ------------------------------------------------------------------
 * function: paint_for_list
 *           paints one process and its childs stored with their 
 *	     positons in pdt without any lines connecting the icons
 * input:    node is the index number of an pdt entry .
 *                This entry and all its subentries will be painted
 *           pointer to xpigc graphic context
 *	     global pdt
 * ------------------------------------------------------------------
 */
PRIVATE void paint_for_list(int node,XpisGC *gc)
{
   int p;
   ProcessIcon *i1,*i2;

   paint_object(pdt[node].picon, gc);
   if((p = pdt[node].cp) != 0) {
      do {
	 paint_for_list(p,gc);
      } while((p=pdt[p].rp) != 0);
   }
}


/* ---------------------------------------------------------------
 * function: make_btree
 *           computes the pointers cp (child pointer) and rp
 *	     (relate pointer) of each entry in pdt to form a 
 *	     binary tree.
 *           input: sorted pdt
 *	     output: filled cp and rp pointers
 * ---------------------------------------------------------------
 */
PRIVATE void make_btree(void)
{
   int i,pid;
   ProcessToDraw *latch;

   /* init all pointers to zero */
   for(i=0;i<(pdt_entries+1);i++) {
      pdt[i].cp = 0;
      pdt[i].rp = 0;
   }

   for(i=1;i<(pdt_entries+1);i++) {
      if((latch = (ProcessToDraw *)bsearch(&pdt[i].ppid,&pdt[1],pdt_entries,
	 sizeof(ProcessToDraw),(fcmp)qscomp)) == NULL)
	 latch = &pdt[0];	 
      pdt[i].rp = latch->cp;
      latch->cp = i;
   }
}


/* ------------------------------------------------------
 * function: qscomp
 *           compare function for bsearch in make_btree
 * ------------------------------------------------------
 */
PRIVATE int qscomp(int *key,ProcessToDraw *elm)
{
   if(*key == elm->pid)
      return(0);
   if(*key <  elm->pid)
      return(-1);
   else
      return(1);
}


/* -------------------------------------------------------------------
 * function: compute_list_pos
 *           computees position of each icon for the list mode
 * input:    gc -> pointer to xpisgc
 *           xdim -> current x dimensions of canvas view window
 * return:   ydim of canvas paint window
 * -------------------------------------------------------------------
 */
PRIVATE void compute_list_pos(int node,int xdim,int *x,int *y,XpisGC *gc)
{
   int p;
   ProcessIcon *icon;

   icon = &pdt[node].picon;
   if(((*x + icon->w) > xdim) && (*x > 0)) {
      *x = 0;
      *y += gc->VerSpareSpace/3 + gc->IconHeight;
   }
   icon->x = *x;
   icon->y = *y;
   *x += icon->w + gc->HorSpareSpace;
   if((p = pdt[node].cp) != 0) {
      do {
	 compute_list_pos(p,xdim,x,y,gc);
      } while((p = pdt[p].rp) != 0);
   }
}


/* -------------------------------------------------------------------
 * function: compute_tree_pos
 *           computes position of each icon for the tree mode
 * input:    gc -> pointer to graphic context
 *           xdim -> the neccessary x dimension will be stored in the
 *                   variable the pointer points to
 *           ydim -> same for y dimension 
 * -------------------------------------------------------------------
 */
PRIVATE void compute_tree_pos(int *xdim,int *ydim,XpisGC *gc)
{
   int maxlevel = 0;
   int minx[MAX_LEVEL];
   int i,*k;

   /* init minx with zero */
   for(i=0;i<MAX_LEVEL;i++)
      minx[i] = 0;

   compute_tree(0,minx,&maxlevel,gc);

   /*look for the largest x dimension*/
   for(k=minx;k<&minx[MAX_LEVEL];k++) {
      if(*k > *xdim)
	 *xdim = *k;
      if(*k == 0)
	 break;
   }

   /*compute ydim using maxlevel*/
   *ydim = ((maxlevel)*(gc->IconHeight+gc->VerSpareSpace));
}


/* -------------------------------------------------------------------
 * function: compute_tree
 *           This function computes the position of node and its child nodes.
 *	     The function calls itself recursively.
 * input:    node -> index number of process in pdt
 *           minx -> pointer to an arry to store  the minimal x position
 *                   of each y-level
 *           maxlevel -> pointer to an integer to store the highst
 *                       reach level
 *           gc -> pointer to xpisgc
 *           global pdt
 * output:   x and y position written in pdt[node].picon.x,y
 * return value: x position of node
 * -------------------------------------------------------------------
 */
PRIVATE int compute_tree(int node,int minx[],int *maxlevel,XpisGC *gc)
{
   static int level = -1;
   ProcessIcon *icon;
   int midpos1, midpos2;
   int p,pt;

  /*set tree level of current node */
   if((++level) >= MAX_LEVEL) {
      puterr("compute_tree: minx[] to small: programm terminated\n");
      exit(-1);
   }

   icon = &pdt[node].picon;
   p = pdt[node].cp;
   if(p == 0)
      icon->x = minx[level];
   else {
      midpos1 = compute_tree(p,minx,maxlevel,gc) + (pdt[p].picon.w/2);
      if(pdt[p].rp == 0)
	 midpos2 = midpos1;
      else {
	 while((p=pdt[pt=p].rp) != 0)
	    midpos2 = compute_tree(p,minx,maxlevel,gc);
         midpos2 += (pdt[pt].picon.w/2);
      }
      icon->x = (midpos2+midpos1-icon->w)/2;
      if(icon->x < minx[level]) 
	 shift_node(node,minx[level]-icon->x,minx,level,gc);
   }
   icon->y = (level-1)*(gc->IconHeight+gc->VerSpareSpace);
   minx[level] = icon->x + icon->w+gc->HorSpareSpace;

   /*store higest level in maxlevel*/
   if(level > *maxlevel)
      *maxlevel = level;

   level--;
   return(icon->x);
}

      
/* ---------------------------------------------------------
 * function: shift_node()
 *           shift_node() moves the given node and all its 
 *           child nodes the given pixels left.
 * input:    node -> index number of node to move
 *           pixels -> number of pixels the node should be 
 *                     moved left
 *           minx -> pointer to the minx array defined in
 *                   compute_tree()
 *           level -> tree level of node
 *           globale pdt
 * output:   new x positions in the appropriate pdt entries
 * ----------------------------------------------------------
 */
PRIVATE void shift_node(int node,int pixels,int minx[],int level,XpisGC *gc)
{
   int p,p1;

   p = pdt[node].cp;
   if(p != 0) {
      do {
	 shift_node(p,pixels,minx,level+1,gc);
      } while((p=pdt[p1=p].rp) != 0);
      minx[level+1] = pdt[p1].picon.x + pdt[p1].picon.w+gc->HorSpareSpace;
   }
   pdt[node].picon.x += pixels;
}


/* --------------------------------------------------------------------
 * function: set_process_icon()
 *	     computes and sets all parameters of the given 
 *           entry in pdt necessary for painting
 * input:    proc -> pointer to pdt entry
 *           gc -> pointer to xpisgc
 *           global data:
 *		pdt
 * --------------------------------------------------------------------
 */
PRIVATE void set_process_icon(ProcessToDraw *proc, XpisGC *gc)
{
   ProcessIcon *icon = &proc->picon;

   /*set icon text*/
   icon->text = proc->pstate->text;

   /*set width of icon in accordance to textw and ImageWidth*/
   if((icon->textw=XTextWidth(gc->xfontstruct,icon->text,
      strlen(icon->text))) > gc->images->ImageWidth)
      icon->w = icon->textw;
   else
      icon->w = gc->images->ImageWidth;
   icon->h = gc->IconHeight;

   set_image(proc, gc);

   if(proc->pid == get_select())
      icon->inverted = YES;
   else
      icon->inverted = NO;
}


/* ---------------------------------------------------------------------
 * function: set_image()
 *           function sets the image of the given process in 
 *           acordance to its state
 * Input:    proc -> pointer to pdt entry
 *           gc -> pointer to xpisgc
 * ---------------------------------------------------------------------
 */
PRIVATE void set_image(ProcessToDraw *proc, XpisGC *gc)
{
   if(*(proc->pstate->status+1) == 'W') {
      proc->picon.image = gc->images->sleeper;
      proc->picon.color = BLUE;
   }
   else {   
      switch(*(proc->pstate->status)) {
         case 'Z':proc->picon.image = gc->images->zombie;
	          proc->picon.color = PURPLE;
	          break;
         case 'T':proc->picon.image = gc->images->stopper;
	          proc->picon.color = YELLOW;
	          break;
         case 'P':
         case 'D':proc->picon.image = gc->images->waiter;
	          proc->picon.color = CYAN;
	          break;
         case 'I':
         case 'S':proc->picon.image = gc->images->sleeper;
	          proc->picon.color = GREEN;
	          break;
         case 'R':proc->picon.image = gc->images->runner;
	          proc->picon.color = RED;
	          break;
         default: puterr("set_image: unknown state\n");
	          proc->picon.image = gc->images->zombie;
	          proc->picon.color = WHITE;
      }
   }
}


/* ----------------------------------------------------------------------
 * function: paint_object
 *           paint one process icon consisting of an image and the text
 *           to the appropriate position
 * input:    icon -> icon to paint
 *           gc -> pointer to xpisgc
 * ----------------------------------------------------------------------
 */
PUBLIC void paint_object(ProcessIcon icon,XpisGC *gc)
{
   int ximage,xtext;
   int d;

   /*calculate position of image and text*/
   if((d=icon.textw - gc->images->ImageWidth) > 0) {
      ximage = icon.x+(d/2);
      xtext = icon.x;
   }
   else {
      ximage = icon.x;
      xtext = icon.x-(d/2);
   }

   /*draw image */
   if(gc->use_color == YES) {
      if(icon.inverted == NO) {
	 XSetBackground(gc->dpy, gc->gc, gc->color_table[icon.color]);
	 XCopyPlane(gc->dpy, icon.image, gc->xwin, gc->gc, 0, 0,
	 gc->images->ImageWidth, gc->images->ImageHeight, ximage,icon.y,1);
	 XSetBackground(gc->dpy, gc->gc, gc->color_table[WHITE]);
      }
      else {
	 XSetForeground(gc->dpy, gc->invgc, gc->color_table[icon.color]);
	 XCopyPlane(gc->dpy, icon.image, gc->xwin, gc->invgc, 0, 0,
	 gc->images->ImageWidth, gc->images->ImageHeight, ximage,icon.y,1);
	 XSetForeground(gc->dpy, gc->invgc, gc->color_table[WHITE]);
      }
   }
   else {
      if(icon.inverted == NO)
	 XCopyPlane(gc->dpy, icon.image, gc->xwin, gc->gc, 0, 0,
	 gc->images->ImageWidth, gc->images->ImageHeight, ximage,icon.y,1);
      else
	 XCopyPlane(gc->dpy,icon.image, gc->xwin, gc->invgc,0,0,
	 gc->images->ImageWidth, gc->images->ImageHeight, ximage,icon.y,1);
   }

   /*draw text */
   XDrawImageString(gc->dpy, gc->xwin, gc->gc,xtext,
	       icon.y + icon.h - gc->xfontstruct->descent,
	       icon.text, strlen(icon.text));
}


/* ----------------------------------------------------------------------
 * function: item_select
 *	     select loops through the pdt and looks for an item to select.
 *	     If no item matches the given coordinates all other selected
 *	     items will be unselected.
 * input:    x,y -> coordinates of pushed mouse pointer
 *           global pdt
 * ----------------------------------------------------------------------
 */
PUBLIC void item_select(int x, int y)
{
   int i, pid, found_one = NO;
   ProcessIcon *icon;
   XpisGC *gc = getgc();
   ProcessToDraw *latch;

   /*Clear previous selection*/
   pid = get_select();
   if(pid >= 0) {
      if((latch = pdt_search(pid)) != NULL) {
         latch->picon.inverted = NO;
	 paint_object(latch->picon,gc);
      }
      clear_select();
   }

   /*select new process if there is one*/
   for(i=1;i<=pdt_entries;i++) {
      icon = &pdt[i].picon;
      if(POINT_IN_RECT(x, y, icon->x, icon->y, icon->w, gc->IconHeight)) {
	 set_select(pdt[i].pid);
	 icon->inverted = YES;
	 paint_object(*icon,gc);
	 found_one = YES;
	 break;
      }
   }

   /*set properties frame*/
   if(found_one == YES)
      set_pitems_p(&pdt[i]);
   else
      deact_pitems();
}

