/**************************************************************************
 * Module name:   getpst.c (get process status table)			  *
 * Project:       xps (Process State Programm for X			  *
 * written by:    Thomas Breitfeld (breitf@infotech.tu-chemnitz.de)	  *
 * last change:   11. 4. 1994                                             *
 *									  *
 * Copyright (c) 1994 Technical University of Chemnitz-Zwickau		  *
 *		      Department of Electrical Engineering		  *
 *		      Division of Data and Communication Technologies	  *
 *									  *
 * ---------------------------------------------------------------------- *
 * Description:   This is the machine dependent module of the xps         *
 *                project. It should make xps work fine on all single     *
 *                processor architectures running SunOS 4.x.              *
 *                If you want xps to run on an other system, you have to  *
 *                rewrite this module.					  *
 *                The main function is getpst(). It reads the process     *
 *                table as it is defined in proc.h out of the kernel      *
 *                memory and fills the xps process table defined in	  *
 *                pst.h with the partially converted values. This xps     *
 *                process table is the interface to the other modules.    *
 * ---------------------------------------------------------------------- *
 * entry points:  getpst_init()						  *
 *		  getpst()						  *
 **************************************************************************/

#include<sys/signal.h>
#include<sys/types.h>

#define KERNEL
#include<sys/param.h>
#undef KERNEL

#include<stdio.h>
#include<kvm.h>
#include<fcntl.h>
#include<nlist.h>
#include<sys/proc.h>
#include<sys/user.h>
#include<errno.h>

#include"xpis.h"
#include"pst.h"

/* Private function prototypes */
PRIVATE int getkval(unsigned long offset, char *ptr, int size, char *refstr);
PRIVATE void get_tty_string(char *tty, dev_t ttyd);
PRIVATE char *get_wchan_string(char *wchan);
PRIVATE void get_state_string(char *string, struct proc *pp);

/* pointer to kernel device*/
PRIVATE kvm_t *kd;

/* variables to access in the kernel*/
#define NPROC 0
#define PROC  1

PRIVATE struct nlist nlst[] = {
	{ "_nproc" },
	{ "_proc" },
	{ 0 }
};

PRIVATE unsigned long proc;
PRIVATE int nproc;

/* pointer to handle the proc list*/
PRIVATE struct proc *pbase;
PRIVATE int bytes;

/* conversion makro for pages into kb*/
int pagesize;
#define pagetokb(size) ((size) * pagesize)

/* conversion macro for pct using the scale factor in param.h*/
#define pctfloat(p) ((float)(p) / FSCALE)


/* --------------------------------------------------------------
 * function: getpst_init
 *           this function initialices the kernel interface and
 *           gets same data for global use
 * return:   nproc if ok
 *             the return value of nproc is used in other modules
 *             to allocate sufficient memory
 *           -1 if error
 * --------------------------------------------------------------
 */
PUBLIC int getpst_init(void)
{
   int i=0;

   /* open kernel interface */
   if((kd = kvm_open(NULL,NULL,NULL,O_RDONLY,"myps")) == NULL) {
      fprintf(stderr,"myps: kvm_open failed\n");
      return(-1);
   }

   /* get offsets of symbols to access in kernel */
   if((i = kvm_nlist(kd,nlst)) < 0) {
      fprintf(stderr,"myps: nlist failed\n");
      return(-1);
   }
   if(i > 0) {
      fprintf(stderr,"myps: nlist: %d symbols not found !\n");
      return(-1);
   }

   /* get the values out of kmem */
   i += getkval(nlst[PROC].n_value, (char *)(&proc), sizeof(proc), "PROC");
   i += getkval(nlst[NPROC].n_value, (char *)&nproc, sizeof(nproc), "NPROC");
   if(i < 0) {     /*an error occured in one of the getkval calls */
       puterr("getpst_init: getkval failed accessing proc and nproc \n");
       return(-1);
   }

   /* get memory for the process table */
   bytes = nproc * sizeof(struct proc);
   if((pbase = (struct proc *)malloc(bytes)) == NULL) {
      fprintf(stderr,"myps: malloc failed\n");
      return(-1);
   }

   /* get pagesize*/
   pagesize = getpagesize();

   return(nproc);
}


/* --------------------------------------------------------------------------
 * function: getpst
 *           getpst gets the system process table and converts its entries
 *           into the xps process tabble
 * input:    pst -> pointer to xps process table
 * return:   number of pst entries if ok
 *           -1 if error
 * --------------------------------------------------------------------------
 */
PUBLIC int getpst(ProcessState *pst)
{
   struct proc *pp;
   struct user *pu;
   struct sess sess;
   struct ucred ucred;
   struct rusage rusage;
   int i, pst_entries = 0;

   /* get process table */
   if(getkval(proc, (char *)pbase, bytes, "Process Table") < 0) {
      puterr("getpst: failed to get process table\n");
      return(-1);
   }

   for(pp = pbase, i = 0; i < nproc; pp++, i++)
   {
       /* look for all used process slots.
	* their state entry is a value greater than zero
	*/
       if(pp->p_stat != 0)
       {
	/*
	 * get the proc's user area and set text
	 */
	   if((pu = kvm_getu(kd, pp)) == NULL) 
	   { 					/*it's a zombie*/
	      strcpy(pst->text, "|");
	      pst->start = 0;

	      if(pp->p_stat == SWAIT)
		 printf("das war abandond\n");

	      /*
	       * get the proc's rusage structure holding exit information
	       */
	      if(getkval((unsigned long)pp->p_ru,(char *)(&rusage),
	   	      sizeof(rusage),"rusage structure") == 0) {
	         pst->ucputime = rusage.ru_utime.tv_sec;
	         pst->scputime = rusage.ru_stime.tv_sec;
	         pst->minflt = rusage.ru_minflt;
	         pst->majflt = rusage.ru_majflt;
	         pst->nswap = rusage.ru_nswap;
	         pst->inblock = rusage.ru_inblock;
	         pst->oublock = rusage.ru_oublock;
	         pst->msgsnd = rusage.ru_msgsnd;
	         pst->msgrcv = rusage.ru_msgrcv;
	         pst->nsignals = rusage.ru_nsignals;
	         pst->nvcsw = rusage.ru_nvcsw;
	         pst->nivcsw = rusage.ru_nivcsw;
	      }
	      else {
	         pst->ucputime = 0;
	         pst->scputime = 0;
	         pst->minflt = 0;
	         pst->majflt = 0;
	         pst->nswap = 0;
	         pst->inblock =00;
	         pst->oublock = 0;
	         pst->msgsnd = 0;
	         pst->msgrcv = 0;
	         pst->nsignals = 0;
	         pst->nvcsw = 0;
	         pst->nivcsw = 0;
              }

	   }

	   else {
	      /*
	       * set commando name of process
	       */
	      if(*(pu->u_comm) == '\0') {
		 if(pp->p_pid == 0) {
		    strcpy(pst->text, "Swapper");
		 }
		 else if(pp->p_pid == 2) {
		    strcpy(pst->text, "Pager");
		 }
	      }
	      else {
		 strcpy(pst->text, pu->u_comm);
	      }

	      pst->start = pu->u_start.tv_sec;
	      /*
	       * read usage structure
	       */
	      pst->ucputime = pu->u_ru.ru_utime.tv_sec;
	      pst->scputime = pu->u_ru.ru_stime.tv_sec;
	      pst->minflt = pu->u_ru.ru_minflt;
	      pst->majflt = pu->u_ru.ru_majflt;
	      pst->nswap = pu->u_ru.ru_nswap;
	      pst->inblock = pu->u_ru.ru_inblock;
	      pst->oublock = pu->u_ru.ru_oublock;
	      pst->msgsnd = pu->u_ru.ru_msgsnd;
	      pst->msgrcv = pu->u_ru.ru_msgrcv;
	      pst->nsignals = pu->u_ru.ru_nsignals;
	      pst->nvcsw = pu->u_ru.ru_nvcsw;
	      pst->nivcsw = pu->u_ru.ru_nivcsw;
	  } 

	/*
	 * get the proc's session struct
	 */
	   if(getkval((unsigned long)pp->p_sessp,(char *)(&sess),
		   sizeof(sess),"session") == 0) {
	      pst->sid  = sess.s_sid;
	      get_tty_string(pst->tty, sess.s_ttyd);
	   }

	/*
	 * get the proc's ucred struct
	 */
	   if(getkval((unsigned long)pp->p_cred,(char *)(&ucred),
		   sizeof(ucred),"credential structure") == 0) {
	      pst->uid  = ucred.cr_uid;
	      pst->gid  = ucred.cr_gid;
	      pst->ruid = ucred.cr_ruid;
	      pst->rgid = ucred.cr_rgid;
	   }

	/*
	 * transfer all other values into pst
	 */
	   pst->pid = pp->p_pid;
	   pst->ppid = pp->p_ppid;
	   pst->pgrp = pp->p_pgrp;
	   pst->pri = pp->p_pri - PZERO;
	   pst->nice = pp->p_nice - NZERO;
	   pst->restime = pp->p_time;
	   pst->slptime = pp->p_slptime;
	   pst->pctcpu = 100 * pctfloat(pp->p_pctcpu);
	   pst->tsize = pagetokb(pp->p_tsize);
	   pst->dsize = pagetokb(pp->p_dsize);
	   pst->ssize = pagetokb(pp->p_ssize);
	   pst->rss = pagetokb(pp->p_rssize);
	   pst->sigignore = pp->p_sigignore;
	   pst->sigcatch = pp->p_sigcatch;
	   pst->flag = pp->p_flag;
	   strcpy(pst->wchan, get_wchan_string(pp->p_wchan));
	   get_state_string(pst->status, pp);

	/*
	 * increase counter and pst pointer
	 */
	   pst_entries++;
	   pst++;
       }
   }
   return(pst_entries);
}


/* -------------------------------------------------------------------
 * function: get_state_string
 * input:    string -> pointer to status string
 *           pp -> pointer to proc structure
 *           slptime -> time since started to sleep ( used
 * output:   characters in string 
 *           first character stands for the state:
 *                 S - sleeping
 *                 R - running
 *                 D - waiting
 *                 I - ideling
 *                 Z - zombie
 *                 T - stoped
 *           second character is set to W if process is 
 *           swapped
 * ------------------------------------------------------------------
 */
PRIVATE void get_state_string(char *string, struct proc *pp)
{
   static char status[] = { '\0', 'S', 'D', 'R', 'R', 'Z', 'T' };

   if((pp->p_stat > 0) && (pp->p_stat <= SSTOP)) {

      *string = status[pp->p_stat];

      if(pp->p_stat == SSLEEP)
      {
         if(pp->p_pri < PZERO)
	    *string = 'D';
         else if(pp->p_slptime > 20)
   	    *string = 'I';
      }	

      if((pp->p_flag & SLOAD) != SLOAD)
         *(++string) = 'W';

      *(++string) = '\0';

   } else {
      fprintf(stderr,"get_state_string error: called with state=%d\n"
		     "Program terminated\n",pp->p_stat);
      exit(1);
   }
}



/* ---------------------------------------------------------------------
 * function:get_wchan_string 
 *          transforms the waitchanel adress to an textstring
 * input:   wchan   
 * return:  pointer to textstring
 * ---------------------------------------------------------------------
 */
PRIVATE char *get_wchan_string(char *wchan)
{
   static char *waitch[] = { "dont know"
			   };

   return(waitch[0]);
}


/* --------------------------------------------------------
 * function: get_tty_string
 * input:    pointer to tty string in pst
 *           tty device id from session struct
 * output:   appropriate string
 *           if unknown device id the id as string
 * --------------------------------------------------------
 */
PRIVATE void get_tty_string(char *tty, dev_t ttyd)
{
   if((ttyd & 0xff00) == 0x1400) {
      *(tty++) = 'p';
      ttyd = ttyd & 0x00ff;
      if(ttyd < 10)
	 *(tty++) = ttyd + '0';
      else
	 *(tty++) = ttyd - 10 + 'a';
      *tty = '\0';
   }
   else if(ttyd == 256) {
      *(tty++) = 'c';
      *(tty++) = 'o';
      *tty = '\0';
   }
   else if(ttyd == 0) {
      *(tty++) = '?';
      *tty = '\0';
   }
   else
      *tty = '\0';
}


/* --------------------------------------------------------------------------
 * function: getkval
 *           get a value out of the kernel.
 * input:    offset -> is the byte offset into the kernel for the desired
 *		       value
 *  	     ptr ->    points to a buffer into which the value is retrieved
 *  	     size ->   is the size of the buffer (and the object to retrieve)
 * return:   0 if ok
 *           -1 if error
 * --------------------------------------------------------------------------
 */
PRIVATE int getkval(unsigned long offset, char *ptr, int size, char *refstr)
{
    if (kvm_read(kd, offset, ptr, size) != size) {
       fprintf(stderr,"xps->getkval: kvm_read for %s failed\n", refstr);
       return(-1);
    }
    return(0);
}
